#  Copyright (c) 2022 by Amplo.

import copy
import time
import warnings
from datetime import datetime
from typing import Dict, Union

import numpy as np
import optuna
import pandas as pd

from amplo.grid_search._base import BaseGridSearch

__all__ = ["OptunaGridSearch"]


class OptunaGridSearch(BaseGridSearch):
    """
    Wrapper for ``optuna`` grid search.

    Takes any model supported by `Amplo.AutoML.Modelling` whose parameter
    search space is predefined for each model.
    Optimal choice [1]

    [1] https://arxiv.org/pdf/2201.06433.pdf

    Parameters
    ----------
    model : Amplo.AutoML.Modeller.ModelType
        Model object to optimize.
    n_trials : int
        Limit the number of trials/candidates to search.
    timeout : int
        Limit the time for optimization.
    cv : sklearn.model_selection.BaseCrossValidator
        Cross validation object.
    scoring : str or sklearn.metrics._scorer._BaseScorer
        A valid string for `sklearn.metrics.get_scorer`.
    verbose : int
        Verbose logging.
    """

    def __init__(self, model, *args, **kwargs):
        if "params" in kwargs:
            warnings.warn("Parameter `params` has no effect")
            kwargs.pop("params")

        super().__init__(model, *args, **kwargs)

        # Counter to keep track of number of trials
        self.trial_count = -1

    @staticmethod
    def _get_suggestion(trial, p_name, p_value):
        """
        Get suggestion for next hyperparameter.

        Parameters
        ----------
        trial : optuna.Trial
            The trial to sample a suggestion from it.
        p_name : str
            Name of the hyperparameter.
        p_value : (str, list) or (str, list, int)
            Sampling information for hyperparameter (dtype, range).

        Returns
        -------
        None or bool or int or float or str
            A specific parameter for given hyperparameter.
        """

        # Read out
        p_type = p_value[0]  # parameter type (str)
        p_args = p_value[1]  # parameter arguments (list, tuple)
        p_kwargs = {}  # additional kwargs parameter, depends on parameter type

        # Sanity checks
        assert (
            len(p_args) == 2 or p_type == "categorical"
        ), "Only categorical parameter can have more/less than two suggest args"

        # Find trial.suggest_{...} type
        if p_type == "categorical":
            p_args = [p_args]  # enclose since will be expanded afterwards
            suggest = trial.suggest_categorical
        elif p_type == "int":
            suggest = trial.suggest_int
        elif p_type == "logint":
            p_kwargs["log"] = True
            suggest = trial.suggest_int
        elif p_type == "uniform":
            suggest = trial.suggest_float
        elif p_type == "loguniform":
            p_kwargs["log"] = True
            suggest = trial.suggest_float
        else:
            raise NotImplementedError("Invalid parameter specification")

        # Suggest parameter given the arguments
        return suggest(p_name, *p_args)

    def _get_hyper_params(self, trial) -> Dict[str, Union[None, bool, int, float, str]]:
        """Use trial to sample from available grid search parameters

        Parameters
        ----------
        trial : optuna.Trial
            The trial from optuna study.

        Returns
        -------
        params : dict of {str : None or bool or int or float or str}
            Sampled grid search parameters.
        """

        param_values = self._hyper_parameter_values
        conditionals = param_values.pop("CONDITIONALS", {})
        params = {}

        # Add non-conditional parameter suggestions (don't depend on another)
        for name, value in param_values.items():
            params[name] = self._get_suggestion(trial, name, value)

        # Add conditional parameter suggestions that satisfy condition:
        for check_p_name, check_p_criteria in conditionals.items():
            for matching_value, additional_params in check_p_criteria:
                # Add parameters only if criterion is satisfied
                if matching_value != params.get(check_p_name, None):
                    # No match
                    continue
                for name, value in additional_params.items():
                    params[name] = self._get_suggestion(trial, name, value)

        return params

    def fit(self, x, y):
        if isinstance(y, pd.DataFrame):
            assert len(y.keys()) == 1, "Multiple target columns not supported."
            y = y[y.keys()[0]]
        assert isinstance(x, pd.DataFrame), "X should be Pandas DataFrame"
        assert isinstance(y, pd.Series), "Y should be Pandas Series or DataFrame"

        # Set mode
        self.binary = y.nunique() == 2
        self.samples = len(y)

        # Store
        self.x, self.y = x, y

        # Set up study
        study = optuna.create_study(
            sampler=optuna.samplers.TPESampler(seed=236868),
            direction="maximize",
            pruner=_BadTrialPruner(2.0, 15),
        )
        study.optimize(
            self.objective,
            timeout=self.timeout,
            n_trials=self.n_trials,
            callbacks=[
                _StopStudyWhenConsecutivePruning(10),
                _StopStudyAfterNPruning(50),
            ],
        )

        # Parse results
        optuna_results = study.trials_dataframe()
        results = pd.DataFrame(
            {
                "date": datetime.today().strftime("%d %b %y"),
                "model": type(self.model).__name__,
                "params": [x.params for x in study.get_trials()],
                "mean_objective": optuna_results["value"],
                "std_objective": optuna_results["user_attrs_std_value"],
                "worst_case": optuna_results["value"]
                - optuna_results["user_attrs_std_value"],
                "mean_time": optuna_results["user_attrs_mean_time"],
                "std_time": optuna_results["user_attrs_std_time"],
            }
        )
        self.trial_count = len(study.trials)

        return results

    def objective(self, trial: optuna.Trial):
        # Metrics
        scores = []
        times = []
        master = copy.deepcopy(self.model)

        # Cross validation
        for step, (t, v) in enumerate(self.cv.split(self.x, self.y)):
            # Split data
            xt, yt = self.x.iloc[t], self.y.iloc[t]
            xv, yv = self.x.iloc[v], self.y.iloc[v]

            # Train model
            t_start = time.time()
            model = copy.deepcopy(master)
            model.set_params(**self._get_hyper_params(trial))
            model.fit(xt, yt)

            # Results
            score = self.scoring(model, xv, yv)
            scores.append(score)
            times.append(time.time() - t_start)

            # Report intermediate values
            trial.report(score, step)

            # Stop cross validation when it is not promising
            if trial.should_prune():
                break

        # Set manual metrics
        mean_score = np.mean(scores)
        trial.set_user_attr("mean_time", np.mean(times))
        trial.set_user_attr("std_time", np.std(times))
        trial.set_user_attr("mean_value", np.mean(scores))
        trial.set_user_attr("std_value", np.std(scores))

        # Stop trial (avoid overwriting)
        if trial.number == self.n_trials:
            trial.study.stop()

        # Pruning
        if trial.should_prune():
            raise optuna.exceptions.TrialPruned

        return mean_score


# ----------------------------------------------------------------------
# Pruning


class _BadTrialPruner(optuna.pruners.BasePruner):
    """
    Pruner to detect outlying metrics of the trials.

    Prune if the mean intermediate value is worse than the best trial value minus (or
    plus) a multiple of its intermediate value standard deviation.

    Parameters
    ----------
    std_threshold_multiplier : float
        Multiplier for the best trials intermediate value std to define the pruning
        threshold.
    n_startup_trials : int
        Pruning is disabled until the given number of trials finish in the same study.
    """

    def __init__(self, std_threshold_multiplier, n_startup_trials):
        if n_startup_trials < 0:
            raise ValueError(
                f"Number of startup trials cannot be negative but got "
                f"{n_startup_trials}."
            )

        self._std_threshold_multiplier = std_threshold_multiplier
        self._n_startup_trials = n_startup_trials

    def prune(self, study, trial):
        # Don't prune while startup trials
        all_trials = study.get_trials(deepcopy=False)
        n_trials = len(
            [t for t in all_trials if t.state == optuna.trial.TrialState.COMPLETE]
        )
        if n_trials < self._n_startup_trials:
            return False

        # Define pruning thresholds
        best_trial_value = study.best_trial.value
        best_trial_value_std = np.std(
            list(study.best_trial.intermediate_values.values())
        )
        threshold = (
            best_trial_value - self._std_threshold_multiplier * best_trial_value_std
        )

        # Pruning
        curr_trial_mean = np.mean(list(trial.intermediate_values.values()))
        if study.direction == optuna.study.StudyDirection.MAXIMIZE:
            return curr_trial_mean < threshold
        else:
            raise RuntimeError(
                "Pruning with a threshold is arbitrary when the study direction is "
                "undefined."
            )


class _StopStudyWhenConsecutivePruning:
    """
    Optuna study callback that stops the study when trials keep being pruned.

    Parameters
    ----------
    threshold : int
        Critical threshold for consecutively pruned trials. Stops trial when achieved.
    """

    def __init__(self, threshold):
        if threshold < 0:
            raise ValueError(f"Threshold cannot be negative but got {threshold}.")

        self._threshold = threshold
        self._consecutive_pruned_count = 0

    def __call__(self, study, trial):
        """
        Stops study when consecutive prune count exceeds threshold.

        Callback function that gets called after each trial.

        Parameters
        ----------
        study : optuna.study.Study
            Study object of the target study.
        trial : optuna.trial.FrozenTrial
            FrozenTrial object of the target trial. Take a copy before modifying this
            object.
        """
        if trial.state == optuna.trial.TrialState.PRUNED:
            self._consecutive_pruned_count += 1
        else:
            self._consecutive_pruned_count = 0

        if self._consecutive_pruned_count > self._threshold:
            study.stop()


class _StopStudyAfterNPruning:
    """
    Optuna study callback that stops the study after begin N times pruned.

    Parameters
    ----------
    threshold : int
        Critical threshold for total number of pruned trials. Stops trial when achieved.
    """

    def __init__(self, threshold):
        if threshold < 0:
            raise ValueError(f"Threshold cannot be negative but got {threshold}.")

        self._threshold = threshold

    def __call__(self, study, trial):
        """
        Stops study when total prune count exceeds threshold.

        Callback function that gets called after each trial.

        Parameters
        ----------
        study : optuna.study.Study
            Study object of the target study.
        trial : optuna.trial.FrozenTrial
            FrozenTrial object of the target trial. Take a copy before modifying this
            object.
        """
        all_trials = study.get_trials(deepcopy=False)
        n_prunings = len(
            [t for t in all_trials if t.state == optuna.trial.TrialState.PRUNED]
        )

        if n_prunings > self._threshold:
            study.stop()



#  Copyright (c) 2022 by Amplo.

import json

import numpy as np
import pandas as pd
import pytest
import pywt

from amplo.automl.feature_processing.pooling import get_pool_functions
from amplo.automl.feature_processing.temporal_feature_extractor import (
    ScoreWatcher,
    TemporalFeatureExtractor,
    _extract_wavelets,
    pl_pool,
)


class TestScoreWatcher:
    def test_singular_scores(self):
        watcher = ScoreWatcher(["a", "b"])
        watcher.update("a", 1)
        watcher.update("a", np.array(9))
        watcher.update("a", 15, weight=2)
        watcher.update("b", 20)
        assert watcher.mean() == 15.0
        assert watcher.std() == 5.0

    def test_multi_scores(self):
        watcher = ScoreWatcher(["a", "b"])
        watcher.update("a", [1, 2, 3])
        watcher.update("a", np.array([9, 8, 7]))
        watcher.update("a", [15, 15, 15], weight=2)
        watcher.update("b", [20, 20, 20])
        assert all(watcher.mean() == 15.0)
        assert all(watcher.std() == 5.0)


@pytest.mark.usefixtures("make_rng")
class TestFunctions:
    def test_pool_single_index(self):
        size = 100
        window_size = 10
        x = pd.Series(self.rng.normal(size=size))

        agg_func = get_pool_functions("mean")

        pooled = pl_pool(x, window_size, agg_func, use_multi_index=False)
        desired_pool = x.values.reshape((-1, window_size)).mean(1)
        assert np.allclose(pooled.values.reshape(-1), desired_pool)

    def test_extract_wavelets(self):
        size = 100
        scales = [1, 10]
        wavelet = "gaus2"

        x = pd.Series(self.rng.normal(size=size))

        transformed = _extract_wavelets(x, scales, wavelet)
        desired_trsf = pywt.cwt(x, scales, wavelet)[0].real.T

        assert np.allclose(transformed.values, desired_trsf)


@pytest.mark.usefixtures("make_rng")
class TestTemporalFeatureExtractor:
    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_mode_and_settings(self, mode, make_x_y):
        x, y = make_x_y
        x = x.iloc[:, :5]  # for speed up
        index = pd.MultiIndex.from_product([[0, 1], range(len(x) // 2)])
        x.index = index
        y.index = index

        fe = TemporalFeatureExtractor(mode=mode, timeout=3)

        # Test output
        out1 = fe.fit_transform(x, y)
        out2 = fe.transform(x)
        assert set(out1) == set(fe.features_), "`features_` doesn't match output."
        assert all(out1 == out2), "`fit_transform` and `transform` don't match."

        # Test settings
        new_fe = TemporalFeatureExtractor().load_settings(fe.get_settings())
        new_out = new_fe.transform(x)
        assert all(out1 == new_out), "FE loaded from settings has invalid output."
        assert set(fe.features_) == set(
            new_fe.features_
        ), "FE from settings has erroneous `features_`."

        # Test JSON serializable
        settings = json.loads(json.dumps(fe.get_settings()))
        new_fe = TemporalFeatureExtractor().load_settings(settings)
        assert fe.get_settings() == new_fe.get_settings()
        assert all(fe.transform(x) == new_fe.transform(x))

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_raw_features(self, mode):
        size = 600
        window_size = 3
        index = pd.MultiIndex.from_product([[0, 1], range(size // 2)])
        if mode == "classification":
            y_np = np.array([*np.zeros(size // 2), *np.ones(size // 2)], dtype=int)
        elif mode == "regression":
            y_np = np.arange(size)
        else:
            raise ValueError("Invalid mode.")

        # Make mean feature.
        #  The mean of each couple of three should be equal to y_pooled
        random1 = self.rng.geometric(0.5, size // window_size)
        random2 = self.rng.normal(size=size // window_size)
        # Formula for mean feature: y_pooled = (random1 + means + random2) / window_size
        means = window_size * y_np[1::window_size] - random1 - random2
        mean_feat = np.array(list(zip(random1, means, random2))).reshape(-1)

        # Make random feature
        random_feat = self.rng.geometric(0.5, size)

        # Convert to pandas
        y = pd.Series(y_np, index=index)
        x = pd.DataFrame(
            {"mean_feat": mean_feat, "random_feat": random_feat}, index=index
        )

        # Setup feature extractor
        fe = TemporalFeatureExtractor(mode=mode)
        fe._set_validation_model()
        fe.window_size_ = window_size
        fe._baseline_scores = [0.99]

        # Fit and check
        y_pooled = fe._pool_target(y)
        fe._fit_transform_raw_features(x, y_pooled)
        assert set(fe.features_) == set(fe.raw_features_)
        assert "mean_feat__pool=mean" in fe.features_, "Mean feature not found."
        assert "random_feat" not in " ".join(fe.features_), "Random feature accepted."

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_wav_features(self, mode):
        # Init
        size = 600
        window_size = 3
        index = pd.MultiIndex.from_product([[0, 1], range(size // 2)])
        fit_wavelets = ["gaus3", "cmor1.0-1.5", "mexh"]

        # Make wavelet contours
        wavelet1 = fit_wavelets[0]
        wavelet2 = fit_wavelets[1]
        wav_contour1, _ = pywt.ContinuousWavelet(wavelet1).wavefun(level=6)  # noqa
        wav_contour2, _ = pywt.ContinuousWavelet(wavelet2).wavefun(level=5)  # noqa
        wav_contour1 = wav_contour1.real
        wav_contour2 = wav_contour2.real

        # Make features
        wav_feat = [
            *np.resize(wav_contour1, size // 2),
            *np.resize(wav_contour2, size // 2),
        ]
        random_feat = self.rng.laplace(size=size)

        # Make target
        if mode == "classification":
            # Make one cluster per log
            y_np = np.array([*np.zeros(size // 2), *np.ones(size // 2)], dtype=int)
        elif mode == "regression":
            # Wavelet transform wav_feat
            y_raw1 = pd.Series(wav_feat, index=index).iloc[: size // 2]
            y_raw2 = pd.Series(wav_feat, index=index).iloc[size // 2 :]
            y_np1 = _extract_wavelets(y_raw1, scales=[16.0], wavelet=wavelet1).values
            y_np2 = _extract_wavelets(y_raw2, scales=[8.0], wavelet=wavelet2).values
            y_np = np.concatenate([y_np1, y_np2]).reshape(-1)
        else:
            raise ValueError("Invalid mode.")

        # Convert to pandas
        y = pd.Series(y_np, index=index)
        x = pd.DataFrame(
            {"wav_feat": wav_feat, "random_feat": random_feat}, index=index
        )

        # Setup feature extractor
        fe = TemporalFeatureExtractor(mode=mode, fit_wavelets=fit_wavelets)
        fe._set_validation_model()
        if mode == "classification":
            fe._baseline_scores = [0.95]
            fe.window_size_ = window_size
        else:
            fe._baseline_scores = [0.7]
            fe.window_size_ = 1

        # Fit and check
        y_pooled = fe._pool_target(y)
        fe._fit_transform_wav_features(x, y_pooled)

        for wav in [wavelet1, wavelet2]:
            assert wav in str(fe.features_), f"Wavelet feature {wav} not found."
        assert "wav_feat" in str(fe.features_), "Didn't accept any wavelet feature."
        assert "random_feat" not in str(fe.features_), "Accepted random feature."

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    @pytest.mark.parametrize("index_type", ["homogeneous", "heterogeneous"])
    def test_set_window_size(self, mode, index_type):
        if index_type == "homogeneous":
            # 10 logs with each 1000 indices
            idx_tuple = [(i, j) for i in range(10) for j in range(100000)]
        elif index_type == "heterogeneous":
            # 10 logs with random number of indices
            idx_tuple = [
                (i, j)
                for i in range(10)
                for j in range(np.random.randint(100, 100_000))
            ]
        else:
            raise ValueError("Invalid index type.")

        index = pd.MultiIndex.from_tuples(idx_tuple)

        # Set window size
        fe = TemporalFeatureExtractor(mode=mode)
        fe._set_window_size(index)

        # Check window size
        if mode == "classification":
            assert fe.window_size_ >= 1
        elif mode == "regression":
            assert fe.window_size_ == 1
        else:
            raise ValueError("Invalid mode.")

    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_pool_target(self, mode, make_x_y):
        size = 90
        fe = TemporalFeatureExtractor(mode=mode)
        fe.window_size_ = 9

        if mode == "classification":
            y_np = self.rng.choice([0, 1], size=size)
            desired_pool = (
                pd.DataFrame(y_np.reshape((-1, fe.window_size_)))
                .apply(lambda row: row.value_counts().idxmax(), axis=1)
                .values
            )
        elif mode == "regression":
            y_np = self.rng.normal(size=size)
            desired_pool = (
                pd.DataFrame(y_np.reshape((-1, fe.window_size_))).mean(axis=1).values
            )
        else:
            raise ValueError("Invalid mode.")

        # Make target and pool
        index = pd.MultiIndex.from_product([[0], range(size)])
        y = pd.Series(y_np, index=index)
        y_pooled = fe._pool_target(y).values

        assert np.allclose(y_pooled, desired_pool), "Pooling doesn't work as expected."

    def test_pool_features(self):
        size = 90
        index = pd.MultiIndex.from_product([[0], range(size)])
        x = pd.DataFrame({"feat_1": self.rng.normal(size=size)}, index=index)

        fe = TemporalFeatureExtractor()
        fe.window_size_ = 9

        # Test instruction not given (make all valid pools)
        out = fe._pool_features(x, instruction=None)
        assert out.shape[0] == size // fe.window_size_

        # Test specific instruction
        out = fe._pool_features(x, instruction={"feat_1": ["mean"]})
        assert out.shape[0] == size // fe.window_size_
        desired_out = x.values.reshape((-1, fe.window_size_, x.shape[1])).mean(1)
        assert np.allclose(out.values, desired_out)

    @pytest.mark.parametrize("n_idx_lvl", range(1, 4))
    def test_check_index(self, n_idx_lvl):
        # Setup
        size = 100
        if n_idx_lvl == 1:
            index = pd.RangeIndex(size)
        elif n_idx_lvl == 2:
            index = pd.MultiIndex.from_product([[0], range(size)])
        elif n_idx_lvl == 3:
            index = pd.MultiIndex.from_product([[0], [0], range(size)])
        else:
            raise ValueError("Invalid parameter 'n_idx_lvl'.")
        x = pd.DataFrame({"feat": np.arange(size)}, index=index)

        # Test data checking
        fe = TemporalFeatureExtractor()
        if n_idx_lvl in (1, 3):
            with pytest.raises(ValueError):
                # Should raise an error when receiving invalid multi-index.
                fe._check_x(x)
        else:
            x_check = fe._check_x(x)
            assert np.allclose(x, x_check)

        # Test convert_single_index option
        if n_idx_lvl == 1:
            x_check = fe._check_x(x, convert_single_index=True)
            assert np.allclose(x, x_check)



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import multiprocessing as mp
import re
from abc import abstractmethod
from typing import Any
from warnings import warn

import numpy as np
import pandas as pd
from sklearn.metrics import get_scorer
from sklearn.metrics._scorer import _BaseScorer  # noqa
from sklearn.model_selection import KFold

from amplo.base.objects import LoggingMixin
from amplo.utils import check_dtypes

__all__ = ["BaseGridSearch"]


class BaseGridSearch(LoggingMixin):
    """
    Abstract base class for grid search.

    Purposes:
        - Enforces to inheriting classes to implement all abstract methods.
        - Defines the hyperparameter search space as it's the same for all
          grid search methods.

    Parameters
    ----------
    model : amplo.automl.modelling.ModelType
        Model object to optimize.
    params : optional
        Parameters to optimize. Has no effect for `OptunaGridSearch`.
    n_trials : int
        Limit the number of trials/candidates to search.
    timeout : int
        Limit the time for optimization.
    cv : sklearn.model_selection.BaseCrossValidator
        Cross validation object.
    scoring : str or sklearn.metrics._scorer._BaseScorer
        A valid string for `sklearn.metrics.get_scorer`.
    verbose : int
        Verbose logging.
    """

    def __init__(
        self,
        model,
        params=None,
        n_trials=250,
        timeout=None,
        cv=KFold(n_splits=10),
        scoring="accuracy",
        verbose=0,
    ):
        super().__init__(verbose=verbose)

        # Input tests
        if hasattr(model, "is_fitted") and model.is_fitted:
            warn(
                "The model is already fitted but Amplo's grid search will re-fit.",
                UserWarning,
            )
        scoring = get_scorer(scoring)
        check_dtypes("scoring", scoring, _BaseScorer)

        # Set class attributes
        self.model = model
        self.params = params
        self.n_trials = n_trials
        self.timeout = timeout
        self.cv = cv
        self.scoring = get_scorer(scoring)

        self.x, self.y = None, None
        self.binary = True
        self.samples = None

        # Model specific settings
        if type(self.model).__name__ == "LinearRegression":
            self.n_trials = 1

    @property
    def _hyper_parameter_values(self) -> dict[str, tuple]:
        """Get model specific hyper parameter values, indicating predefined
        search areas to optimize.

        Notes
        -----
        Each item of the output dictionary consists of:
            - parameter name (str), and
            - parameter specifications (tuple)
                - parameter type (str)
                - parameter arguments (list)
                - number of distinct values (int, optional) [only for exhaustive grid
                  search]

        Parameter types include:
            - 'categorical': categorical values
            - 'int': discretized uniform value space
            - 'logint': discretized logarithmic uniform value space
            - 'uniform': uniform value space
            - 'loguniform': logarithmic uniform value space

        Parameter arguments are:
            - [categorical]: a list of all options
            - [int, logint, uniform, loguniform]: a tuple with min and max value

        **Special case (conditionals):**
            In some cases, one wants to grid-search certain parameters only
            if another parameter condition is present. Such conditions are
            specified via the dedicated key 'CONDITIONALS'.
        """

        # Extract model name & type
        model_name = type(self.model).__name__
        model_type = re.split(r"Regressor|Classifier", model_name)[0]
        # Determine whether it's classification or regression
        is_regression = bool(re.match(r".*(Regression|Regressor|SVR)", model_name))
        is_classification = bool(
            re.match(r".*(Classification|Classifier|SVC)", model_name)
        )
        if not (is_regression or is_classification):
            raise ValueError("Could not determine mode (regression or classification).")

        # Define min-max-function
        def minimax(min_, value, max_):
            return max(min_, min(value, max_))

        # Find matching model and return its parameter values

        if model_name == "LinearRegression" or model_name == "LogisticRegression":
            return {}

        elif model_name == "Lasso" or "Ridge" in model_name:
            return dict(alpha=("uniform", [0, 10], 25))

        elif model_name in ("SVR", "SVC"):
            return dict(
                gamma=("categorical", ["scale", "auto", 0.001, 0.01, 0.1, 0.5, 1]),
                C=("uniform", [0.001, 10], 25),
            )

        elif model_type == "KNeighbors":
            return dict(
                n_neighbors=("int", [5, minimax(5, self.samples // 10, 50)], 5),
                weights=("categorical", ["uniform", "distance"]),
                leaf_size=("int", [1, minimax(1, self.samples // 10, 100)], 5),
                n_jobs=("categorical", [mp.cpu_count() - 1]),
            )

        elif model_type == "MLP":
            raise NotImplementedError("MLP is not supported")

        elif model_type == "SGD":
            params = dict(
                loss=(
                    "categorical",
                    [
                        "squared_loss",
                        "huber",
                        "epsilon_insensitive",
                        "squared_epsilon_insensitive",
                    ],
                ),
                penalty=("categorical", ["l2", "l1", "elasticnet"]),
                alpha=("uniform", [0, 10], 5),
                max_iter=("int", [250, 1000], 3),
            )
            if is_classification:
                params.update(
                    loss=(
                        "categorical",
                        ["hinge", "log", "modified_huber", "squared_hinge"],
                    )
                )
            return params

        elif model_type == "DecisionTree":
            params = dict(
                criterion=(
                    "categorical",
                    ["squared_error", "friedman_mse", "absolute_error", "poisson"],
                ),
                max_depth=("int", [3, minimax(3, int(np.log2(self.samples)), 25)], 4),
            )
            if is_classification:
                params.update(criterion=("categorical", ["gini", "entropy"]))
            return params

        elif model_type == "AdaBoost":
            params = dict(
                n_estimators=("int", [25, 250], 5),
                loss=("categorical", ["linear", "square", "exponential"]),
                learning_rate=("loguniform", [0.001, 1], 10),
            )
            if is_classification:
                params.pop("loss", None)
            return params

        elif model_type == "Bagging":
            params = dict(
                n_estimators=("int", [10, 250], 4),
                max_samples=("uniform", [0.5, 1.0], 4),
                max_features=("uniform", [0.5, 1.0], 4),
                n_jobs=("categorical", [mp.cpu_count() - 1]),
            )
            return params

        elif model_type == "CatBoost":
            params = dict(
                n_estimators=("int", [500, 2000], 5),
                verbose=("categorical", [0]),
                od_pval=("categorical", [1e-5]),
                loss_function=("categorical", ["MAE", "RMSE"]),
                learning_rate=("loguniform", [0.001, 0.5], 5),
                l2_leaf_reg=("uniform", [0, 10], 5),
                depth=("int", [3, minimax(3, int(np.log2(self.samples)), 10)], 4),
                min_data_in_leaf=("int", [1, minimax(1, self.samples // 10, 1000)], 5),
                grow_policy=(
                    "categorical",
                    ["SymmetricTree", "Depthwise", "Lossguide"],
                ),
            )
            if is_classification:
                params.update(
                    loss_function=(
                        "categorical",
                        ["Logloss" if self.binary else "MultiClass"],
                    )
                )
            return params

        elif model_type == "GradientBoosting":
            params = dict(
                loss=("categorical", ["ls", "lad", "huber"]),
                learning_rate=("loguniform", [0.001, 0.5], 10),
                max_depth=("int", [3, minimax(3, int(np.log2(self.samples)), 10)], 4),
                n_estimators=("int", [100, 1000], 4),
                min_samples_leaf=(
                    "int",
                    [1, minimax(1, int(self.samples / 10), 1000)],
                    3,
                ),
                max_features=("uniform", [0.5, 1], 3),
                subsample=("uniform", [0.5, 1], 3),
            )
            if is_classification:
                params.update(loss=("categorical", ["deviance", "exponential"]))
            return params

        elif model_type == "HistGradientBoosting":
            params = dict(
                loss=("categorical", ["least_squares", "least_absolute_deviation"]),
                learning_rate=("loguniform", [0.001, 0.5], 10),
                max_iter=("int", [100, 250], 4),
                max_leaf_nodes=("int", [30, 150], 4),
                max_depth=("int", [3, minimax(3, int(np.log2(self.samples)), 10)], 4),
                min_samples_leaf=(
                    "int",
                    [1, minimax(1, int(self.samples / 10), 1000)],
                    4,
                ),
                l2_regularization=("uniform", [0, 10], 5),
                max_bins=("int", [100, 255], 4),
                early_stopping=("categorical", [True]),
            )
            if is_classification:
                params.pop("loss", None)
            return params

        elif model_type == "RandomForest":
            params = dict(
                n_estimators=("int", [50, 1000], 5),
                criterion=("categorical", ["squared_error", "absolute_error"]),
                max_depth=("int", [2, minimax(3, int(np.log2(self.samples)), 15)], 4),
                max_features=("categorical", ["log2", "sqrt"]),
                min_samples_split=("int", [2, 50], 4),
                min_samples_leaf=("int", [1, minimax(1, self.samples // 10, 1000)], 5),
                bootstrap=("categorical", [True, False]),
            )
            if is_classification:
                params.update(criterion=("categorical", ["gini", "entropy"]))
            return params

        elif model_type == "XGB":
            params = dict(
                objective=("categorical", ["reg:squarederror"]),
                eval_metric=("categorical", ["rmse"]),
                booster=("categorical", ["gbtree", "gblinear", "dart"]),
                alpha=("loguniform", [1e-8, 1], 10),
                learning_rate=("loguniform", [0.001, 0.5], 5),
                n_jobs=("categorical", [mp.cpu_count() - 1]),
            )
            params["lambda"] = ("loguniform", [1e-8, 1], 10)
            if is_classification:
                params.update(
                    # objective=("categorical", ["multi:softprob"]),
                    eval_metric=("categorical", ["logloss"]),
                )
            params["CONDITIONALS"] = dict(
                booster=[
                    (
                        "gbtree",
                        dict(
                            max_depth=(
                                "int",
                                [1, minimax(1, int(np.log2(self.samples)), 10)],
                                5,
                            ),
                            eta=("loguniform", [1e-8, 1], 5),
                            gamma=("loguniform", [1e-8, 1], 5),
                            grow_policy=("categorical", ["depthwise", "lossguide"]),
                        ),
                    ),
                    (
                        "dart",
                        dict(
                            max_depth=(
                                "int",
                                [1, minimax(1, int(np.log2(self.samples)), 10)],
                                5,
                            ),
                            eta=("loguniform", [1e-8, 1], 5),
                            gamma=("loguniform", [1e-8, 1], 5),
                            grow_policy=("categorical", ["depthwise", "lossguide"]),
                            sample_type=("categorical", ["uniform", "weighted"]),
                            normalize_type=("categorical", ["tree", "forest"]),
                            rate_drop=("loguniform", [1e-8, 1], 5),
                            skip_drop=("loguniform", [1e-8, 1], 5),
                        ),
                    ),
                ],
            )
            return params

        elif model_type == "LGBM":
            if is_regression:
                return dict(
                    num_leaves=("int", [10, 150], 5),
                    min_data_in_leaf=(
                        "int",
                        [1, minimax(1, self.samples // 10, 1000)],
                        0,
                    ),
                    min_sum_hessian_in_leaf=("uniform", [0.001, 0.5], 0),
                    colsample_bytree=("uniform", [0, 1], 5),
                    reg_alpha=("uniform", [0, 1], 5),
                    reg_lambda=("uniform", [0, 1], 5),
                    verbosity=("categorical", [-1]),
                    n_jobs=("categorical", [mp.cpu_count() - 1]),
                )
            else:  # is_classification
                return dict(
                    objective=(
                        "categorical",
                        ["binary" if self.binary else "multiclass"],
                    ),
                    metric=(
                        "categorical",
                        ["binary_error", "auc", "average_precision", "binary_logloss"]
                        if self.binary
                        else ["multi_error", "multi_logloss", "auc_mu"],
                    ),
                    boosting_type=("categorical", ["gbdt"]),
                    lambda_l1=("loguniform", [1e-8, 10], 4),
                    lambda_l2=("loguniform", [1e-8, 10], 4),
                    num_leaves=("int", [10, 5000], 4),
                    max_depth=("int", [5, 20], 4),
                    min_data_in_leaf=(
                        "int",
                        [1, minimax(1, self.samples // 10, 1000)],
                        0,
                    ),
                    min_gain_to_split=("uniform", [0, 5], 0),
                    feature_fraction=("uniform", [0.4, 1], 0),
                    bagging_fraction=("uniform", [0.4, 1], 0),
                    bagging_freq=("int", [1, 7], 0),
                    verbosity=("categorical", [-1]),
                    n_jobs=("categorical", [mp.cpu_count() - 1]),
                )

        # Raise error if no match was found
        raise NotImplementedError(
            "Hyper parameter tuning not implemented for {}".format(model_name)
        )

    def get_parameter_min_max(self) -> pd.DataFrame:
        """
        Get all min and max values from model-specific set of parameters.

        Omit categorical parameters as min and max values are ambiguous.

        Returns
        -------
        param_min_max : pd.DataFrame
            Min and max values.
        """

        # Get all model's parameters
        param_values = self._hyper_parameter_values
        # Pop conditionals and integrate all into `param_values`
        conditionals = param_values.pop("CONDITIONALS", {})
        for check_p_name, check_p_criteria in conditionals.items():
            for matching_value, additional_params in check_p_criteria:
                for name, value in additional_params.items():
                    param_values[name] = value

        # Filter for min and max in non-categorical parameters
        param_min_max = {}
        for p_name, value in param_values.items():
            p_type, p_args = value[:2]
            if p_type != "categorical":  # ("int", "logint", "uniform", "loguniform")
                min_, max_ = p_args
                param_min_max[p_name] = {"min": min_, "max": max_}

        # Combine all values to pd.DataFrame
        param_min_max = pd.DataFrame(param_min_max).T
        return param_min_max

    @abstractmethod
    def _get_hyper_params(self, *args, **kwargs) -> dict[str, Any]:
        """
        Get grid search specific distributions or samples.

        This function translates `self._hyper_parameter_values` to the expected
        format for the given grid search.

        Parameters
        ----------
        args : optional
            Grid search specific arguments.
        kwargs : optional
            Grid search specific keyword arguments.

        Returns
        -------
        Grid search parameters
        """
        pass

    @abstractmethod
    def fit(self, x, y) -> pd.DataFrame:
        """
        Run fit with model-specific set of parameters.

        Parameters
        ----------
        x : array
            Data features.
        y : array
            Data labels

        Returns
        -------
        results : pd.DataFrame
            Results of the grid search.
        """
        pass



#  Copyright (c) 2022 by Amplo.

import pytest

from amplo import Pipeline
from amplo.observation import ProductionObserver
from amplo.observation._base import ProductionWarning
from tests import DelayedRandomPredictor


class TestProductionObserver:
    @pytest.mark.parametrize("mode", ["classification", "regression"])
    def test_check_prediction_latency(self, mode, make_x_y):
        pytest.skip()

        x, y = make_x_y

        # Make pipeline and simulate fit
        pipeline = Pipeline(n_grid_searches=0)
        pipeline._read_data(x, y)
        pipeline._mode_detector()
        pipeline.best_model = DelayedRandomPredictor(delay=0.1, mode=mode)
        pipeline.best_model.fit(x, y)

        # Observe
        obs = ProductionObserver(pipeline)
        with pytest.warns(ProductionWarning):
            obs.check_prediction_latency(threshold=0.1)



#  Copyright (c) 2022 by Amplo.

import numpy as np
import pytest

from amplo.automl.modelling import Modeller
from amplo.utils import check_dtypes, clean_feature_name, get_model, hist_search


class TestUtils:
    def test_get_model(self):
        # Test valid models
        for mode in ("classification", "regression"):
            for samples in [100, 100_000]:
                all_models = set(
                    type(model)
                    for model in Modeller(mode=mode, samples=samples).return_models()
                )
                for model_type in all_models:
                    model_name = model_type.__name__
                    model = get_model(model_name, mode=mode, samples=samples)
                    assert type(model) == model_type

        # Test invalid model
        try:
            get_model("ImaginaryModel", mode="regression", samples=100)
            raise AssertionError("`get_model` did not raise a ValueError.")
        except ValueError:
            pass

    def test_hist_search(self):
        bin_idx = hist_search(np.arange(100).astype(float), 50)
        assert bin_idx == 50, "Returned wrong bin index."

    def test_clean_feature_name(self):
        ugly = "   This-is (an)UGLY   [ string__"
        pretty = "this_is_an_ugly_string"
        assert pretty == clean_feature_name(ugly)

    def test_check_dtypes(self):
        # Test valid dtypes
        valid = [("1", 1, int), ("2.0", 2.0, float), ("both", 3.0, (int, float))]
        check_dtypes(*valid)
        # Test invalid dtypes
        invalid = [("invalid", 1.0, int)]
        with pytest.raises(TypeError):
            check_dtypes(*invalid)



#  Copyright (c) 2022 by Amplo.

from catboost import CatBoostRegressor as _CatBoostRegressor
from sklearn.model_selection import train_test_split

from amplo.regression._base import BaseRegressor
from amplo.utils import check_dtypes


class CatBoostRegressor(BaseRegressor):
    """
    Amplo wrapper for catboost.CatBoostRegressor.

    Parameters
    ----------
    test_size : float, default: 0.1
        Test size for train-test-split in fitting the model.
    random_state : int, default: None
        Random state for train-test-split in fitting the model.
    verbose : {0, 1, 2}, default: 0
        Verbose logging.
    **model_params : Any
        Model parameters for underlying catboost.CatBoostRegressor.
    """

    model: _CatBoostRegressor  # type hint

    def __init__(self, test_size=0.1, random_state=None, verbose=0, **model_params):
        # Verify input dtypes and integrity
        check_dtypes(
            ("test_size", test_size, float),
            ("random_state", random_state, (type(None), int)),
            ("model_params", model_params, dict),
        )
        if not 0 <= test_size < 1:
            raise ValueError(f"Invalid attribute for test_size: {test_size}")

        # Set up model
        default_model_params = {
            "n_estimators": 1000,
            "allow_writing_files": False,
            "early_stopping_rounds": 100,
            "use_best_model": True,
            "verbose": verbose,
        }
        for k, v in default_model_params.items():
            if k not in model_params:
                model_params[k] = v
        model = _CatBoostRegressor(**model_params)

        # Set attributes
        self.test_size = test_size
        self.random_state = random_state

        super().__init__(model=model, verbose=verbose)

    def _fit(self, x, y=None, **fit_params):
        # Split data and fit model
        xt, xv, yt, yv = train_test_split(
            x, y, test_size=self.test_size, random_state=self.random_state
        )
        self.model.fit(
            xt,
            yt,
            eval_set=[(xv, yv)],
            early_stopping_rounds=self.model.get_params().get("early_stopping_rounds"),
            use_best_model=self.model.get_params().get("use_best_model"),
        )



#  Copyright (c) 2022 by Amplo.

import requests

from amplo.utils import check_dtypes

__all__ = ["BaseRequestAPI"]


class BaseRequestAPI:
    """
    Base class for API.

    Parameters
    ----------
    host : str
        API host.
    access_token : str
        API access token.
    """

    def __init__(self, host: str, access_token: str):
        check_dtypes(("host", host, str), ("access_token", access_token, str))
        self.host = host
        self.access_token = access_token

    def __repr__(self):
        """
        Readable string representation of the class.
        """

        return f"{self.__class__.__name__}({self.host.removeprefix('https://')})"

    def _authorization_header(self) -> dict:
        raise NotImplementedError("Abstract method.")

    def request(self, method: str, action: str, **kwargs) -> requests.Response:
        """
        Send a request to the API.

        Parameters
        ----------
        method : str
            Request method (``GET``, ``PUT``, ``POST``, ...).
        action : str
            Path to API action.
        kwargs : Any, optional
            Additional keyword arguments for the request (``params``, ``headers``, ...).

        Returns
        -------
        requests.Response
            The request's response.

        Raises
        ------
        requests.HTTPError
            When the request's response has another status code than 200.
        """

        # Add authorization string to "headers" key
        headers: dict = kwargs.get("headers") or {}
        check_dtypes("kwargs__headers", headers, dict)
        headers.update(self._authorization_header())

        # Request
        url = f"{self.host}/api/{action.lstrip('/')}"
        response = requests.request(method, url, headers=headers, **kwargs)

        # Verify response
        if response.status_code != 200:
            raise requests.HTTPError(f"{response} {response.text}")

        return response
